package org.netkernel.module.standard.builtin;

import java.util.*;

import org.netkernel.container.IKernel;
import org.netkernel.layer0.meta.IPrototypeMeta;
import org.netkernel.layer0.meta.IPrototypeParameterMeta;
import org.netkernel.layer0.meta.IRepresentationMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.impl.PrototypeMetaImpl;
import org.netkernel.layer0.meta.impl.PrototypeParameterMetaImpl;
import org.netkernel.layer0.meta.impl.RepresentationMetaImpl;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.StandardModule;
import org.netkernel.module.standard.endpoint.IStandardOverlay;
import org.netkernel.module.standard.endpoint.StandardEndpointImpl;
import org.netkernel.request.IRequest;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;
import org.w3c.dom.Element;
import org.w3c.dom.Node;


public class MetaProviderEndpoint extends StandardEndpointImpl
{
	
	private static final PairList VP = new PairList(5);
	static
	{	VP.put("*[name()!='class' and name()!='id'  and name()!='parameter' and name()!='private' and "+StandardModule.STANDARD_METADATA_PARAMETERS_VALIDATION+"]", "MSG_STD_METAPROV_P_CHILD_TAGS");
		VP.put("self::node()[count(class)!=1]", "MSG_STD_METAPROV_P_NO_IMPL");
		VP.put("parameter/attribute::*[name()!='name' and name()!='desc' and name()!='type' and name()!='default' and name()!='max' and name()!='min']", "MSG_STD_METAPROV_P_PARAM_ATTS");
		VP.put("parameter[ * or text() ]", "MSG_STD_METAPROV_P_PARAM_CONTENT");
		VP.put("parameter[not(@name and @type)]", "MSG_STD_METAPROV_P_PARAM_REQUIRED");
	}
	
	private static final PairList VR = new PairList(4);
	static
	{	VR.put("*[name()!='class' and name()!='id' and name()!='private' and  "+StandardModule.STANDARD_METADATA_PARAMETERS_VALIDATION+"]", "MSG_STD_METAPROV_R_CHILD_TAGS");
		VR.put("self::node()[count(class)!=1]", "MSG_STD_METAPROV_R_NO_IMPL");
	}
	
	private static final Map<String,Class> sParameterTypes = new HashMap();
	static
	{	sParameterTypes.put("space",ISpace.class);
		sParameterTypes.put("element",Element.class);
		sParameterTypes.put("identifier",IIdentifier.class);
		sParameterTypes.put("string",String.class);
		sParameterTypes.put("integer",Integer.class);
		sParameterTypes.put("float",Float.class);
		sParameterTypes.put("long",Long.class);
		sParameterTypes.put("double",Double.class);
		sParameterTypes.put("boolean",Boolean.class);
	}
	
	
	private String mId;
	private IMetaRepresentation mMeta;
	
	//private IEndpointParts mParts;
	//private Map mIdentifierToMeta;
	//private String mToString=getClass().getName();
	
	public MetaProviderEndpoint()
	{	declareThreadSafe();
		declareParameters(MetaProviderPrototype.sPrototypeParameters);
	}
	
	public void onCommissionEndpoint(IKernel aKernel, ISpace aSpace) throws Exception
	{	super.onCommissionEndpoint(aKernel, aSpace);
		//mIdentifierToMeta = new HashMap();
	
		// legacy parsing for classes
		/*
		List partClasses=getParameters().getParameterValues(MetaProviderPrototype.PARAM_CLASS);
		for (Iterator i=partClasses.iterator(); i.hasNext(); )
		{	String partClass = (String)i.next();
			IMetaRepresentation meta=null;
			Class c = aContext.getClassLoader().loadClass(partClass);
			String identifierString;
			if (IRepresentationMeta.class.isAssignableFrom(c))
			{	IRepresentationMeta rmeta = (IRepresentationMeta)c.newInstance();
				identifierString=rmeta.getImplementation().getName();
				meta=rmeta;
			}
			else if (IPrototypeMeta.class.isAssignableFrom(c))
			{	IPrototypeMeta pmeta = (IPrototypeMeta)c.newInstance();
				identifierString = pmeta.getIdentifier().toString();
				meta=pmeta;
			}
			else
			{	IRepresentationMeta rmeta = new DefaultRepresentationMeta(c);
				identifierString=rmeta.getImplementation().getName();
				meta=rmeta;
			}
			
			mIdentifierToMeta.put(identifierString,meta);
		}
		*/
		
		/*
		List<Element> prototypes=getParameters().getParameterValues("prototype");
		for (Element protoElement: prototypes )
		{	parsePrototype(protoElement, mIdentifierToMeta);
		}
		
		List<Element> representations=getParameters().getParameterValues("representation");
		for (Element repElement: representations )
		{	parseRepresentation(repElement, mIdentifierToMeta);
		}
		*/
		Element e = (Element)getParameter("config");
		String name=XMLUtils.pleaseDOMGiveMeTheNameOf(e);
		if (name.equals("representation"))
		{	parseRepresentation(e);
		}
		else if (name.equals("prototype"))
		{	parsePrototype(e);
		}

		/*
		StringBuffer sb=new StringBuffer();
		sb.append("Metadata declaration for ");
		PairList partsList = new PairList(mIdentifierToMeta.size());
		for (Iterator i=mIdentifierToMeta.keySet().iterator(); i.hasNext(); )
		{	String id=(String)i.next();
			sb.append('\"');
			sb.append(id);
			sb.append('\"');
			if (i.hasNext())
			{	sb.append(", ");
			}
			partsList.put(new SimpleIdentifierImpl(id),null);
		}
		mParts = new EndpointPartsImpl(partsList);
		mToString=sb.toString();
		*/
	}
	
	private void parsePrototype(Element aProtoElement) throws NKFException
	{
		PairList validation=FastSchematron.validate(aProtoElement, VP);
		if (validation.size()>0)
		{	String m = FastSchematron.formatValidationError(validation,getKernel().getLogger());			
			throw createParseException("MSG_STD_METAPROV_SCHEMA",null,m);
		}
		
		XMLReadable prototype = new XMLReadable(aProtoElement);
		String classname = prototype.getText("class");
		Class implementationClass=null;
		try
		{	implementationClass = getSpace().getClassLoader().loadClass(classname);
		}
		catch (ClassNotFoundException e)
		{	throw createParseException("MSG_STD_METAPROV_P_BAD_CLASS",aProtoElement,classname);
		}
		String identifierString=prototype.getText("id");
		if (identifierString.length()==0)
		{	identifierString=classname;
		}
		String name=prototype.getText("name");
		if (name.length()==0)
		{	name=Utils.shortClassName(implementationClass);
		}
		String desc=prototype.getText("description");
		if (desc.length()==0)
		{	desc=null;
		}
		List<Element> parameters=(List<Element>)(List)prototype.getNodes("parameter");
		IPrototypeParameterMeta[] params = new IPrototypeParameterMeta[parameters.size()];
		int i=0;
		for (Element parameter: parameters)
		{	// name, type, default, max, min
			String pName=parameter.getAttribute("name");
			
			String pDesc=parameter.getAttribute("desc");
			if (pDesc.length()==0) pDesc=null;
			
			String pTypeString=parameter.getAttribute("type");
			Class pType = sParameterTypes.get(pTypeString);
			if (pType==null)
			{	throw createParseException("MSG_STD_METAPROV_P_BAD_PARAM_TYPE",parameter,pTypeString);
			}
			
			String defaultValueString = parameter.getAttribute("default");
			Object defaultValue;
			if (defaultValueString.length()==0)
			{	defaultValue = IPrototypeParameterMeta.NO_DEFAULT;
			}
			else
			{	if (defaultValueString.equals("null"))
				{	defaultValue=null;
				}
				else if (pType==IStandardOverlay.class || pType==Element.class)
				{	throw createParseException("MSG_STD_METAPROV_P_BAD_DEFAULT",parameter,defaultValueString);
				}
				else if (pType==IIdentifier.class)
				{	defaultValue = new SimpleIdentifierImpl(defaultValueString);
				}
				else if (pType==String.class)
				{	defaultValue = defaultValueString;
				}
				else if (pType==Integer.class)
				{	defaultValue = Integer.parseInt(defaultValueString);
				}
				else if (pType==Boolean.class)
				{	defaultValue = Boolean.parseBoolean(defaultValueString);
				}
				else if (pType==Long.class)
				{	defaultValue = Long.parseLong(defaultValueString);
				}
				else if (pType==Float.class)
				{	defaultValue = Float.parseFloat(defaultValueString);
				}
				else if (pType==Double.class)
				{	defaultValue = Double.parseDouble(defaultValueString);
				}
				else
				{	//should not happen
					defaultValue = null;
				}
			}
			
			boolean noDefault = defaultValueString.length()==0;
			
			int min=Utils.parseValency(parameter.getAttribute("min"),noDefault?1:0);
			int max=Utils.parseValency(parameter.getAttribute("max"),1);
			
			if (min==-1 || max==-1 || min>max || min==Integer.MAX_VALUE)
			{	throw createParseException("MSG_STD_METAPROV_P_BAD_PARAM_VALENCY",parameter);
			}			
			
			IPrototypeParameterMeta parameterMeta = new PrototypeParameterMetaImpl(pName,pDesc,pType,defaultValue,min,max);
			params[i++]=parameterMeta;
		}
		
		PrototypeMetaImpl meta = new PrototypeMetaImpl(identifierString,implementationClass,name,desc,params);
		
		boolean isPrivate=false;
		Node isPrivateNode=prototype.getFirstNode("private");
		if (isPrivateNode!=null)
		{	isPrivate=!XMLUtils.getText(isPrivateNode).equals("false");
		}
		if (isPrivate)
		{	meta.addAdditionalField(PrivateFilterEndpoint.KEY_PRIVATE, Boolean.TRUE);
		}
		
		mMeta=meta;
		mId=identifierString;
	}
	
	private void parseRepresentation(Element aRepElement) throws NKFException
	{
		PairList validation=FastSchematron.validate(aRepElement, VR);
		if (validation.size()>0)
		{	String m = FastSchematron.formatValidationError(validation,getKernel().getLogger());			
			throw createParseException("MSG_STD_METAPROV_SCHEMA",null,m);
		}
		XMLReadable prototype = new XMLReadable(aRepElement);
		String classname = prototype.getText("class");
		Class implementationClass=null;
		try
		{	implementationClass = getSpace().getClassLoader().loadClass(classname);
		}
		catch (ClassNotFoundException e)
		{	throw createParseException("MSG_STD_METAPROV_R_BAD_CLASS",aRepElement,classname);
		}
		String identifierString=prototype.getText("id");
		if (identifierString.length()==0)
		{	identifierString=classname;
		}
		String name=prototype.getText("name");
		if (name.length()==0)
		{	name=Utils.shortClassName(implementationClass);
		}
		String desc=prototype.getText("description");
		if (desc.length()==0)
		{	desc=null;
		}
		RepresentationMetaImpl meta = new RepresentationMetaImpl(implementationClass,name,desc);
		
		boolean isPrivate=false;
		Node isPrivateNode=prototype.getFirstNode("private");
		if (isPrivateNode!=null)
		{	isPrivate=!XMLUtils.getText(isPrivateNode).equals("false");
		}
		if (isPrivate)
		{	meta.addAdditionalField("private", Boolean.TRUE);
		}
		
		mMeta=meta;
		mId=identifierString;
	}
	
	
	private NKFException createParseException(String aMessageId, Element aLocation, Object... aMessageParams)
	{	return Utils.createFormattedNKFException("EX_STD_METAPROV_CONFIG", aMessageId, XMLUtils.getPathFor(aLocation), null, getKernel().getLogger(), aMessageParams);
	}
	
	
	public ISpaceElements getParts(INKFRequestContext aContext) throws NKFException
	{	TupleList p = new TupleList(3,1);
		p.put(new SimpleIdentifierImpl(mId),null,this);
		return new SpaceElementsImpl(p);
	}

	public void onResolve(INKFResolutionContext aContext) throws NKFException
	{	// doesn't resolve anything
	}
	
	public void onMeta(INKFRequestContext aContext) throws NKFException
	{	IIdentifier id=(IIdentifier)aContext.getKernelContext().getThisKernelRequest().getValue(IRequest.TYPE_RESOLVED_ENDPOINT_IDENTIFIER);
		if (id.equals(mId))
		{	aContext.createResponseFrom(mMeta);
		}
	}
	
	public String toString()
	{	return mId;
	}
}
